/*
 *
 *  Copyright (C) 1996, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Forschungsbereich 2: Kommunikationssysteme
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  for CEN/TC251/WG4 as a contribution to the Computer Assisted Radiology
 *  (CAR) 1996 DICOM Demonstration.
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Copyright of the software  and  supporting  documentation  is,  unless
 *  otherwise stated, owned by OFFIS, and free access is hereby granted as
 *  a license to  use  this  software,  copy  this  software  and  prepare
 *  derivative works based upon this software.  However, any  distribution
 *  of this software source code or supporting documentation or derivative
 *  works  (source code and  supporting documentation)  must  include  the
 *  three paragraphs of this copyright notice.
 *
 *
 * Module: DICOM Basic Worklist Management CTN, WWW Component
 *
 * Author: Andrew Hewett
 *
 * Purpose:
 *   Program to create a worklist file from a WWW CGI perl script generated 
 *   hexedecimal encoded string.
 *
 * Last Update:      $Author: meichel $
 * Update Date:      $Date: 1997/05/30 07:33:28 $
 * Source File:      $Source: /share/dicom/cvs-depot/dcmtk/wlistctn/wwwapps/writwlst.cc,v $
 * CVS/RCS Revision: $Revision: 1.6 $
 * Status:           $State: Exp $
 *
 * CVS/RCS Log
 *   $Log: writwlst.cc,v $
 *   Revision 1.6  1997/05/30 07:33:28  meichel
 *   Added space characters around comments and simplified
 *   some inlining code (needed for SunCC 2.0.1).
 *
 *   Revision 1.5  1997/05/29 17:06:45  meichel
 *   All dcmtk applications now contain a version string
 *   which is displayed with the command line options ("usage" message)
 *   and which can be queried in the binary with the "ident" command.
 *
 *   Revision 1.4  1997/05/22 13:39:38  hewett
 *   Modified the test for presence of a data dictionary to use the
 *   method DcmDataDictionary::isDictionaryLoaded().
 *
 *   Revision 1.3  1997/04/18 08:48:10  andreas
 *   - The put/get-methods for all VRs did not conform to the C++-Standard
 *     draft. Some Compilers (e.g. SUN-C++ Compiler, Metroworks
 *     CodeWarrier, etc.) create many warnings concerning the hiding of
 *     overloaded get methods in all derived classes of DcmElement.
 *     So the interface of all value representation classes in the
 *     library are changed rapidly, e.g.
 *     E_Condition get(Uint16 & value, const unsigned long pos);
 *     becomes
 *     E_Condition getUint16(Uint16 & value, const unsigned long pos);
 *     All (retired) "returntype get(...)" methods are deleted.
 *     For more information see dcmdata/include/dcelem.h
 *
 *   Revision 1.2  1996/06/10 09:04:13  meichel
 *   Removed static const variable in class definition. Sun CC does not
 *   support this.
 *
 *   Revision 1.1  1996/05/31 11:43:36  meichel
 *   Initial release of the C/C++ code for the Worklist CTN WWW/CGI application.
 *
 *
 */

#include "osconfig.h"
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <ctype.h>
#include "dctk.h"
#include "dcdebug.h"
#include "dcuid.h"    /* for dcmtk version name */

static char rcsid[] = "$dcmtk: writwlst v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";


// ********************************************

static void
usage()
{
    cerr << rcsid << "\n\n"
	"writlst: write a worklist file from a hex coded string\n"
	"usage: wlistgen [options] hexstring-file-in dcmfile-out\n"
	"arguments:\n"
	"  hexstring-file-in   a filename of - reads from stdin\n"
	"  dcmfile-out         the dicom file to create\n"
	"options are:\n"
	"  create options:\n"
	"    +f    write fileformat (default)\n"
	"    -f    write dataset\n"
	"  group length encoding:\n" 
	"    +g    write with group lengths\n"
	"    -g    write without group lengths (default)\n"
	"  length encoding in sequences and items:\n"
	"    +e    write with explicit lengths (default)\n"
	"    -e    write with undefined lengths\n"
	"  output transfer syntax:\n"
	"    +ti   write with little-endian implicit transfer syntax\n"
	"    +te   write with little-endian explicit transfer syntax (default)\n"
	"    +tb   write with big-endian explicit transfer syntax\n"
	"  other test/debug options:\n"
	"    +V    verbose mode, print actions\n"
	"    +v    validate input data (currently almost useless)\n"
	"    +dn   set debug level to n (n=1..9)\n";
}


// ********************************************

#define SimpleStr_defaultSize 32

class SimpleStr {
private:
    char* str;
    int len;
    int maxlen;

    int max(int a, int b) {
	return ((a)>(b))?(a):(b);
    }
    void expand(int newlen);
public:
    int length() const { 
	return len; 
    }
    int maxlength() const { 
	return maxlen; 
    }

    const char* c_str() const { 
	return str; 
    }

    void append(char c) {
	expand(len+1);
	str[len] = c;
	len++;
	str[len] = '\0';
    }
    void append(const char* s) {
	if (s) {
	    int slen = strlen(s);
	    expand(len+slen);
	    strcat(str, s);
	    len += slen;
	}
    }
    void append(const SimpleStr& s) {
	append(s.c_str());
    }

    void clear() {
	if (str) {
	    str[0] = '\0';
	    len = 0;
	}
    }

    char& operator[](int i) {
	return str[i];
    }

    SimpleStr(int capacity = SimpleStr_defaultSize) { 
	str = NULL;
	len = maxlen = 0;
	expand(capacity);
    }
    SimpleStr(const char* s) {
	str = NULL;
	len = maxlen = 0;
	append(s);
    }
    SimpleStr(const SimpleStr& s) {
	str = NULL;
	len = maxlen = 0;
	append(s);
    }
    ~SimpleStr() { delete str; }

};

void SimpleStr::expand(int newlen) 
{
    if (newlen >= maxlen) {
        maxlen = max(newlen, maxlen+SimpleStr_defaultSize);
        char* newstr = new char[maxlen+1];
        newstr[0] = '\0';
        if (str) {
            strcpy(newstr, str);
            delete str;
        }
        str = newstr;
    }
}

// ********************************************

static BOOL
addStringAttr(DcmItem *item, const DcmTagKey& key, const SimpleStr& str)
{
    E_Condition ec = EC_Normal;
    DcmTag tag(key);
    DcmElement *newElement = NULL;

    if (item == NULL) {
	cerr << "error: cannot add to null item" << endl;
	return FALSE;
    }

    ec = newDicomElement(newElement, tag);
    if (ec != EC_Normal) {
	cerr << "error: cannot create element: " << key << endl;
	return FALSE;
    }

    ec = newElement->putString(str.c_str());
    if (ec != EC_Normal) {
	cerr << "error: cannot put element value: " << key << endl;
	return FALSE;
    }

    ec = item->insert(newElement);
    if (ec != EC_Normal) {
	cerr << "error: cannot insert element: " << key << endl;
	return FALSE;
    }

    return TRUE;
}


static DcmItem*
getOrInsertSequenceItem(DcmItem *dset, const DcmTagKey& key, 
			unsigned int itemNumber)
{
    E_Condition ec = EC_Normal;
    DcmStack stack;
    DcmItem *foundItem = NULL;

    if (dset == NULL) {
	cerr << "error: cannot add to null item" << endl;
	return NULL;
    }

    ec = dset->search(key, stack, ESM_fromHere, FALSE);

    if (stack.empty()) {
	/* insert the sequence and items */
	DcmTag tag(key);
	DcmSequenceOfItems *sq = new DcmSequenceOfItems(tag);
	for (unsigned int i=0; i<=itemNumber; i++) {
	    DcmItem *item = new DcmItem;
	    sq->insert(item);
	}
	dset->insert(sq);
	foundItem = sq->getItem(itemNumber);
    } else if (stack.top()->ident() == EVR_SQ) {
	DcmSequenceOfItems *sq = (DcmSequenceOfItems*)stack.top();
	if (itemNumber > sq->card()) {
	    /* we need to create some more items */ 
	    for (unsigned long i=sq->card(); i<=itemNumber; i++) {
		DcmItem *item = new DcmItem;
		sq->insert(item);
	    }
	} 
	foundItem = sq->getItem(itemNumber);
    } else {
	cerr << "error: not a sequence: " << key << endl;
	return NULL;
    }

    return foundItem;
}

// ********************************************

struct KeyPos {
    int position;
    DcmTagKey key;
    BOOL withinSequence;

    KeyPos(int pos, const DcmTagKey& k, BOOL inSeq = FALSE);
};

KeyPos::KeyPos(int pos, const DcmTagKey& k, BOOL inSeq)
{
    position = pos;
    key = k;
    withinSequence = inSeq;
}

// ********************************************

static BOOL
addValue(int valCount, SimpleStr& val, DcmDataset& dset)
{
    static KeyPos posTab[] = {
	KeyPos(  0, DCM_RequestedProcedureID ),
	KeyPos(  1, DCM_RequestedProcedureDescription ),
	KeyPos(  2, DCM_StudyInstanceUID ),
	KeyPos(  3, DCM_RequestedProcedurePriority ),
	KeyPos(  4, DCM_AccessionNumber ),
	KeyPos(  5, DCM_RequestingPhysician ),
	KeyPos(  6, DCM_PatientName ),
	KeyPos(  7, DCM_PatientID ),
	KeyPos(  8, DCM_PatientBirthDate ),
	KeyPos(  9, DCM_PatientSex ),
	KeyPos( 10, DCM_MedicalAlerts ),
	KeyPos( 11, DCM_ContrastAllergies ),
	KeyPos( 12, DCM_ScheduledStationAETitle, TRUE ),
	KeyPos( 13, DCM_ScheduledProcedureStepStartDate, TRUE ),
	KeyPos( 14, DCM_ScheduledProcedureStepStartTime, TRUE ),
	KeyPos( 15, DCM_Modality, TRUE ),
	KeyPos( 16, DCM_ScheduledPerformingPhysiciansName, TRUE ),
	KeyPos( 17, DCM_ScheduledProcedureStepDescription, TRUE ),
	KeyPos( 18, DCM_ScheduledStationName, TRUE ),
	KeyPos( 19, DCM_ScheduledProcedureStepLocation, TRUE ),
	KeyPos( 20, DCM_PreMedication, TRUE ),
	KeyPos( 21, DCM_ScheduledProcedureStepID, TRUE ),
	KeyPos( 22, DCM_RequestedContrastAgent, TRUE ),
	KeyPos( 23, DCM_CommentsOnTheScheduledProcedureStep, TRUE )
    };

    int posTabCount = (sizeof(posTab) / sizeof(posTab[0]));
    if (valCount < 0 || valCount >= posTabCount) {
	/* ignore extra values */
	return TRUE;
    }
    
    KeyPos *kp = &posTab[valCount];
    if (kp->position != valCount) {
	cerr << "internal error: inconsistant posTab (valCount=" 
	     << valCount <<")" << endl;
	return FALSE;
    }

#ifdef DEBUG
    printf("adding string: %s\n", val.c_str());
#endif

    BOOL ok = TRUE;
    if (kp->withinSequence) {
	DcmItem* item1 = getOrInsertSequenceItem(
	    &dset, DCM_ScheduledProcedureStepSequence, 0);
	ok = addStringAttr(item1, kp->key, val);
    } else {
	ok = addStringAttr(&dset, kp->key, val);
    }
    return ok;
}

// ********************************************

#define SEPARATOR_CHAR '\\'

static BOOL
parseWorklist(FILE* f, DcmDataset& dset)
{
    SimpleStr hexChar(2);
    SimpleStr val;
    int valCount = 0;
    int c;

    /* ignore first '\\' */
    while (((c = fgetc(f)) != EOF) && isspace(c)) {
	/* do nothing */
    }
    if (c != SEPARATOR_CHAR) {
	cerr << "error: expected \'" << char(SEPARATOR_CHAR) << "\' " <<
	    "as first input character" << endl;
	return FALSE;
    }

    while ((c = fgetc(f)) != EOF) {
	if (c == SEPARATOR_CHAR) {
	    addValue(valCount, val, dset);
	    val.clear();
	    valCount++;
	} else if (isxdigit(c)) {
	    hexChar.append(c);
	    if (hexChar.length() == 2) {
		/* convert into normal character */
		unsigned int h;
		const char *s = hexChar.c_str();
		sscanf(s, "%x", &h);
		char newc = (h & 0xff);
		val.append(newc);
		hexChar.clear();
	    }
	} else if (isspace(c)) {
	    /* ignore */
	} else {
	    cerr << "error bad hex value: " << char(c) << endl;
	    return FALSE;
	}
    }

    /* add any trailing value */
    addValue(valCount, val, dset);

    return TRUE;
}

// ********************************************

int main(int argc, char *argv[])
{
    SetDebugLevel(( 0 ));

    if (argc < 3) {
	usage();
        return 1;
    }

    const char *ifname = NULL;
    const char *ofname = NULL;
    DcmFileFormat *fileformat = NULL;
    DcmDataset *dataset = NULL;
    E_TransferSyntax xfer = EXS_LittleEndianExplicit;
    E_EncodingType enctype = EET_ExplicitLength;
    E_GrpLenEncoding ogltype = EGL_withoutGL;
    BOOL verifymode = FALSE;
    BOOL verbosemode = FALSE;
    BOOL createFileFormat = TRUE;
    int localDebugLevel = 0;

    for (int i=1; i<argc; i++) {
	char* arg = argv[i];
	if ((strcmp(arg, "-") == 0) && (ifname == NULL)) {
	    ifname = arg;
	} else if (arg[0] == '-' || arg[0] == '+') {
	    if (strlen(arg) < 2) {
		cerr << "unknown argument: " << arg << endl;
		usage();
		return 1;
	    }
	    switch (arg[1]) {
	    case 'f':
		if (arg[0] == '-') {
		    createFileFormat = FALSE;
		} else {
		    createFileFormat = TRUE;
		}
		break;
	    case 'g':
		if (arg[0] == '-') {
		    ogltype = EGL_withoutGL;
		} else {
		    ogltype = EGL_withGL;
		}
		break;
	    case 'e':
		if (arg[0] == '-') {
		    enctype = EET_UndefinedLength;
		} else {
		    enctype = EET_ExplicitLength;
		}
		break;
	    case 't':
		switch (arg[2]) {
		case 'i':
		    xfer = EXS_LittleEndianImplicit;
		    break;
		case 'e':
		    xfer = EXS_LittleEndianExplicit;
		    break;
		case 'b':
		    xfer = EXS_BigEndianExplicit;
		    break;
		default:
		    cerr << "unknown option: " << arg << endl;
		    return 1;
		}
		break;
	    case 'v':
		if (arg[0] == '+' && arg[2] == '\0') {
		    verifymode = TRUE;
		} else {
		    cerr << "unknown option: " << arg << endl;
		    return 1;
		}
		break;
	    case 'V':
		if (arg[0] == '+' && arg[2] == '\0') {
		    verbosemode = TRUE;
		} else {
		    cerr << "unknown option: " << arg << endl;
		    return 1;
		}
		break;
	    case 'd':
		if (sscanf(arg+2, "%d", &localDebugLevel) != 1) {
		    cerr << "unknown option: " << arg << endl;
		    return 1;
		}
		break;
	    default:
		cerr << "unknown option: " << arg << endl;
		return 1;
	    }
	} else if ( ifname == NULL ) {
	    ifname = arg;
	} else if ( ofname == NULL ) {
	    ofname = arg;
	} else {
	    cerr << "too many arguments: " << arg << endl;
	    usage();
	    return 1;
	}
    }

    if ( ifname == NULL ) {
	cerr << "missing input file\n";
	return 1;
    }

    if ( ofname == NULL ) {
	cerr << "missing output file\n";
	return 1;
    }

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
	cerr << "Warning: no data dictionary loaded, "
	     << "check environment variable: "
	     << DCM_DICT_ENVIRONMENT_VARIABLE << endl;
    }
    
    SetDebugLevel(( localDebugLevel ));

    // create dicom metaheader and dataset
    BOOL memoryError = FALSE;
    if (createFileFormat) {
	fileformat = new DcmFileFormat();
	if (fileformat)	{
	    dataset = fileformat -> getDataset();
	    if (!dataset)
		memoryError = TRUE;
	} else {
	    memoryError = TRUE;
	}
    } else {
	dataset = new DcmDataset();
	if (!dataset)
	    memoryError = TRUE;
    }

    if (memoryError) {
	cerr << "virtual memory exhausted" << endl;
	return 1;
    }

    BOOL parseOK = TRUE;

    if (strcmp(ifname, "-") == 0) {
	/* read from stdin */
	parseOK = parseWorklist(stdin, *dataset);
	
    } else {
	FILE* f = NULL;
	if ((f = fopen(ifname, "r")) != NULL) {
	    parseOK = parseWorklist(f, *dataset);
	    fclose(f);
	} else {
	    cerr << "cannot open file: " << ifname << endl;
	}
    }

    if (!parseOK) {
	return 1;
    }

    /* must add an attribute for the specific character set */
    addStringAttr(dataset, DCM_SpecificCharacterSet, "ISO_IR 100");

    /* verify dataset */

    if (verifymode) {
	if (verbosemode)
	    cout << "verifying dataset" << endl;
	dataset->verify(TRUE);
    }

    /* write to file */
    
    if (verbosemode)
	cout << "writing file: " << ofname << endl;

    DcmFileStream oStream(ofname, DCM_WriteMode);

    if (oStream.Fail())	{
	cerr << "cannot open output file: " << ofname << endl;
	return 1;
    }

    E_Condition l_error = EC_Normal;
    if (fileformat) {
	fileformat -> transferInit();
	l_error = fileformat -> write(oStream, xfer, enctype, ogltype);
    } else if (dataset)	{
	dataset -> transferInit();
	l_error = dataset -> write(oStream, xfer, enctype, ogltype);
    } else {
	cerr << "internal error: no fileform or dataset" << endl;
	return 1;
    }

    if (l_error != EC_Normal) {
	cerr << "Error: " << dcmErrorConditionToString(l_error) 
	    << ": writing file: "  << ofname << endl;
	return 1;
    }

    return 0;
}
